==> ns-1.0b1.tar.gz: ns-1.0b1/tcp-newreno.cc <==
/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: tcp-newreno.cc,v 1.4 96/01/27 15:40:41 mccanne Exp $ (LBL)";
#endif

//
// newreno-tcp: a revised reno TCP source, without sack
//
// first cut, Nov 1995; not yet debugged
//

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"

class NewRenoTcpAgent : public TcpAgent {
 public:
	NewRenoTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
 protected:
	u_int dupwnd_;
	void partialnewack(Packet *pkt);
};

static class NewRenoTcpMatcher : public Matcher {
public:
	NewRenoTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno") == 0)
			return (new NewRenoTcpAgent());
		return (0);
	}
} matcher_reno;

int NewRenoTcpAgent::window()
{
	int win = int(cwnd_) + dupwnd_;
	if (win > int(2 * wnd_))
		win = int(2 * wnd_);
	return (win);
}

NewRenoTcpAgent::NewRenoTcpAgent() : dupwnd_(0)
{
}

/* 
 * Process a packet that acks previously unacknowleges data, but
 * does not take us out of Fast Retransmit.
 */
void NewRenoTcpAgent::partialnewack(Packet* pkt)
{
        newtimer(pkt);
#ifdef notyet
        if (pkt->seqno_ == stp->maxpkts && stp->maxpkts > 0)
                stp->endtime = (float) realtime();
#endif
        last_ack_ = pkt->seqno_;
        highest_ack_ = last_ack_;
        if (seqno_ < last_ack_ + 1)
                seqno_ = last_ack_ + 1;
        if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
                rtt_active_ = 0;
                t_backoff_ = 1;
        }
}

void NewRenoTcpAgent::recv(Packet *pkt)
{
	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (pkt->seqno_ > last_ack_) {
	    if (pkt->seqno_ >= recover_) {
		dupwnd_ = 0;
		newack(pkt);
                opencwnd();
	    }
	    else {
		partialnewack(pkt);
		output(last_ack_ + 1, 0);
	    }
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
			dupwnd_ = NUMDUPACKS;
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				closecwnd(1);
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
		} else if (dupacks_ > NUMDUPACKS)
			++dupwnd_;
	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0) 
		send(0, 0, maxburst_);
	else if (dupacks_ > NUMDUPACKS - 1)
		send(0, 0, 2);
}

void NewRenoTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
		TcpAgent::timeout(tno);
	}
}

==> ns-1.1.tar.gz: ns-1.1/tcp-newreno.cc <==
/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char rcsid[] =
"@(#) $Header: tcp-newreno.cc,v 1.2 96/11/11 22:43:37 floyd Exp $ (LBL)";
#endif

//
// newreno-tcp: a revised reno TCP source, without sack
//

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>

#include "tcp.h"

class NewRenoTcpAgent : public TcpAgent {
 public:
	NewRenoTcpAgent();
	virtual int window();
	virtual void recv(Packet *pkt);
	virtual void timeout(int tno);
 protected:
	u_int dupwnd_;
	int changes_;		/* 0 for fixing unnecessary fast retransmits */
				/* 1 for additional code from Allman, */
				/*  to implement other algorithms from */
				/*  Hoe's paper */
	void partialnewack(Packet *pkt);
};

static class NewRenoTcpMatcher : public Matcher {
public:
	NewRenoTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-newreno") == 0)
			return (new NewRenoTcpAgent());
		return (0);
	}
} matcher_reno;

int NewRenoTcpAgent::window()
{
        //
        // reno: inflate the window by dupwnd_
        //      dupwnd_ will be non-zero during fast recovery,
        //      at which time it contains the number of dup acks
        //
	int win = int(cwnd_) + dupwnd_;
        if (win > int(wnd_))
                win = int(wnd_);
        return (win);
}

NewRenoTcpAgent::NewRenoTcpAgent() : dupwnd_(0), changes_(0)
{
	link_int("ns_tcpnewreno", "changes", &changes_, 0);
}

/* 
 * Process a packet that acks previously unacknowleges data, but
 * does not take us out of Fast Retransmit.
 */
void NewRenoTcpAgent::partialnewack(Packet* pkt)
{
        newtimer(pkt);
#ifdef notyet
        if (pkt->seqno_ == stp->maxpkts && stp->maxpkts > 0)
                stp->endtime = (float) realtime();
#endif
        last_ack_ = pkt->seqno_;
        highest_ack_ = last_ack_;
        if (seqno_ < last_ack_ + 1)
                seqno_ = last_ack_ + 1;
        if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
                rtt_active_ = 0;
                t_backoff_ = 1;
        }
}

void NewRenoTcpAgent::recv(Packet *pkt)
{
	static int acked = 0;
	int new_ssthresh;
	static double ack2, ack3, basertt;

        /* Use first packet to calculate the RTT  --contributed by Allman */

        if (++acked == 1)
                basertt = Scheduler::instance().clock() - firstsent_;

        /* Estimate ssthresh based on the calculated RTT and the estimated
           bandwidth (using ACKs 2 and 3).  */

        else if (acked == 2)
                ack2 = Scheduler::instance().clock();
        else if (acked == 3)
        {       
                ack3 = Scheduler::instance().clock();  
                new_ssthresh = int((basertt * (size_ / (ack3 - ack2)))
                                    / size_);
                if (changes_ > 0 && new_ssthresh < ssthresh_)
                    ssthresh_ = new_ssthresh;
        }

	if (pkt->type_ != PT_ACK) {
		fprintf(stderr,
			"ns: confiuration error: tcp received non-ack\n");
		exit(1);
	}

	if (pkt->flags_ & PF_ECN)
		quench(1);
	if (pkt->seqno_ > last_ack_) {
	    if (pkt->seqno_ >= recover_) {
		dupwnd_ = 0;
		newack(pkt);
                opencwnd();
	    }
	    else {
		/* received new ack for a packet sent during Fast
		 *  Recovery, but sender stays in Fast Recovery */
		dupwnd_ = 0;
		partialnewack(pkt);
		output(last_ack_ + 1, 0);
	    }
   	} else if (pkt->seqno_ == last_ack_)  {
		if (++dupacks_ == NUMDUPACKS) {
			/*
			 * Assume we dropped just one packet.
			 * Retransmit last ack + 1
			 * and try to resume the sequence.
			 */
                       /* The line below, for "bug_fix_" true, avoids
                        * problems with multiple fast retransmits after
			* a retransmit timeout.
                        */
			if ( (highest_ack_ > recover_) ||
			    ( recover_cause_ != 2)) {
				recover_cause_ = 1;
				recover_ = maxseq_;
				closecwnd(1);
				reset_rtx_timer(1);
				output(last_ack_ + 1, PF_DUPACK);
                        }
			dupwnd_ = NUMDUPACKS;
		} else if (dupacks_ > NUMDUPACKS) 
			++dupwnd_;
                        /* For every two duplicate ACKs we receive (in the
                         * "fast retransmit phase"), send one entirely new
                         * data packet "to keep the flywheel going".  --Allman 
			 */
			 if (changes_ > 0 && (dupacks_ % 2) == 1)
		   		output (seqno_++,0);

	}
	Packet::free(pkt);
	if (trace_)
		plot();

	/*
	 * Try to send more data
	 */

	if (dupacks_ == 0) 
                /*
                 * Maxburst is really only needed for the first
                 *   window of data on exiting Fast Recovery.
                 */
		send(0, 0, maxburst_);
	else if (dupacks_ > NUMDUPACKS - 1 && changes_==0)
		send(0, 0, 2);
}

void NewRenoTcpAgent::timeout(int tno)
{
	if (tno == TCP_TIMER_RTX) {
		dupwnd_ = 0;
		dupacks_ = 0;
		if (bug_fix_) recover_ = maxseq_;
		TcpAgent::timeout(tno);
	}
}
